"""
Sampling
========

Draw samples from the distribution, with support for various samplers.
"""

from .sample import sample
from .sampler import Sampler, InternalSampler
from .metropolis import MetropolisSampler
from .adaptive_metropolis import AdaptiveMetropolisSampler
from .parallel_tempering import ParallelTemperingSampler
from .adaptive_parallel_tempering import AdaptiveParallelTemperingSampler
from .pymc3 import Pymc3Sampler
from .result import McmcPtResult
from .diagnostics import geweke_test
